@@ -4,12 +4,12 @@
  *	  This file contains heap tuple accessor and mutator routines, as well
  *	  as various tuple utilities.
  *
- * Portions Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/access/common/heaptuple.c,v 1.74 2001/10/25 05:49:20 momjian Exp $
+ *	  $Header: /cvsroot/pgsql/src/backend/access/common/heaptuple.c,v 1.87.2.1 2005/02/06 20:15:32 tgl Exp $
  *
  * NOTES
  *	  The old interface functions have been converted to macros
@@ -48,7 +48,7 @@ ComputeDataSize(TupleDesc tupleDesc,
 		if (nulls[i] != ' ')
 			continue;
 
-		data_length = att_align(data_length, att[i]->attlen, att[i]->attalign);
+		data_length = att_align(data_length, att[i]->attalign);
 		data_length = att_addlength(data_length, att[i]->attlen, value[i]);
 	}
 
@@ -69,7 +69,7 @@ DataFill(char *data,
 {
 	bits8	   *bitP = 0;
 	int			bitmask = 0;
-	uint32		data_length;
+	Size		data_length;
 	int			i;
 	int			numberOfAttributes = tupleDesc->natts;
 	Form_pg_attribute *att = tupleDesc->attrs;
@@ -80,7 +80,7 @@ DataFill(char *data,
 		bitmask = CSIGNBIT;
 	}
 
-	*infomask &= HEAP_XACT_MASK;
+	*infomask &= ~(HEAP_HASNULL | HEAP_HASVARWIDTH | HEAP_HASEXTENDED);
 
 	for (i = 0; i < numberOfAttributes; i++)
 	{
@@ -105,17 +105,18 @@ DataFill(char *data,
 		}
 
 		/* XXX we are aligning the pointer itself, not the offset */
-		data = (char *) att_align((long) data, att[i]->attlen, att[i]->attalign);
+		data = (char *) att_align((long) data, att[i]->attalign);
 
 		if (att[i]->attbyval)
 		{
 			/* pass-by-value */
 			store_att_byval(data, value[i], att[i]->attlen);
+			data_length = att[i]->attlen;
 		}
 		else if (att[i]->attlen == -1)
 		{
 			/* varlena */
-			*infomask |= HEAP_HASVARLENA;
+			*infomask |= HEAP_HASVARWIDTH;
 			if (VARATT_IS_EXTERNAL(value[i]))
 				*infomask |= HEAP_HASEXTERNAL;
 			if (VARATT_IS_COMPRESSED(value[i]))
@@ -123,15 +124,22 @@ DataFill(char *data,
 			data_length = VARATT_SIZE(DatumGetPointer(value[i]));
 			memcpy(data, DatumGetPointer(value[i]), data_length);
 		}
+		else if (att[i]->attlen == -2)
+		{
+			/* cstring */
+			*infomask |= HEAP_HASVARWIDTH;
+			data_length = strlen(DatumGetCString(value[i])) + 1;
+			memcpy(data, DatumGetPointer(value[i]), data_length);
+		}
 		else
 		{
 			/* fixed-length pass-by-reference */
-			Assert(att[i]->attlen >= 0);
-			memcpy(data, DatumGetPointer(value[i]),
-				   (size_t) (att[i]->attlen));
+			Assert(att[i]->attlen > 0);
+			data_length = att[i]->attlen;
+			memcpy(data, DatumGetPointer(value[i]), data_length);
 		}
 
-		data = (char *) att_addlength((long) data, att[i]->attlen, value[i]);
+		data += data_length;
 	}
 }
 
@@ -165,13 +173,11 @@ heap_attisnull(HeapTuple tup, int attnum)
 			case MinCommandIdAttributeNumber:
 			case MaxTransactionIdAttributeNumber:
 			case MaxCommandIdAttributeNumber:
+				/* these are never null */
 				break;
 
-			case 0:
-				elog(ERROR, "heap_attisnull: zero attnum disallowed");
-
 			default:
-				elog(ERROR, "heap_attisnull: undefined negative attnum");
+				elog(ERROR, "invalid attnum: %d", attnum);
 		}
 
 	return 0;
@@ -222,9 +228,9 @@ nocachegetattr(HeapTuple tuple,
 	/* ----------------
 	 *	 Three cases:
 	 *
-	 *	 1: No nulls and no variable length attributes.
-	 *	 2: Has a null or a varlena AFTER att.
-	 *	 3: Has nulls or varlenas BEFORE att.
+	 *	 1: No nulls and no variable-width attributes.
+	 *	 2: Has a null or a var-width AFTER att.
+	 *	 3: Has nulls or var-widths BEFORE att.
 	 * ----------------
 	 */
 
@@ -235,7 +241,8 @@ nocachegetattr(HeapTuple tuple,
 		if (att[attnum]->attcacheoff != -1)
 		{
 			return fetchatt(att[attnum],
-				  (char *) tup + tup->t_hoff + att[attnum]->attcacheoff);
+							(char *) tup + tup->t_hoff +
+							att[attnum]->attcacheoff);
 		}
 #endif
 	}
@@ -243,9 +250,7 @@ nocachegetattr(HeapTuple tuple,
 	{
 		/*
 		 * there's a null somewhere in the tuple
-		 */
-
-		/*
+		 *
 		 * check to see if desired att is null
 		 */
 
@@ -319,9 +324,9 @@ nocachegetattr(HeapTuple tuple,
 
 	/*
 	 * If slow is false, and we got here, we know that we have a tuple
-	 * with no nulls or varlenas before the target attribute. If possible,
-	 * we also want to initialize the remainder of the attribute cached
-	 * offset values.
+	 * with no nulls or var-widths before the target attribute. If
+	 * possible, we also want to initialize the remainder of the attribute
+	 * cached offset values.
 	 */
 	if (!slow)
 	{
@@ -346,11 +351,7 @@ nocachegetattr(HeapTuple tuple,
 			  (HeapTupleNoNulls(tuple) || !att_isnull(j, bp)) &&
 			  (HeapTupleAllFixed(tuple) || att[j]->attlen > 0)); j++)
 		{
-			/*
-			 * Fix me when going to a machine with more than a four-byte
-			 * word!
-			 */
-			off = att_align(off, att[j]->attlen, att[j]->attalign);
+			off = att_align(off, att[j]->attalign);
 
 			att[j]->attcacheoff = off;
 
@@ -391,7 +392,7 @@ nocachegetattr(HeapTuple tuple,
 				off = att[i]->attcacheoff;
 			else
 			{
-				off = att_align(off, att[i]->attlen, att[i]->attalign);
+				off = att_align(off, att[i]->attalign);
 
 				if (usecache)
 					att[i]->attcacheoff = off;
@@ -399,11 +400,11 @@ nocachegetattr(HeapTuple tuple,
 
 			off = att_addlength(off, att[i]->attlen, tp + off);
 
-			if (usecache && att[i]->attlen == -1)
+			if (usecache && att[i]->attlen <= 0)
 				usecache = false;
 		}
 
-		off = att_align(off, att[attnum]->attlen, att[attnum]->attalign);
+		off = att_align(off, att[attnum]->attalign);
 
 		return fetchatt(att[attnum], tp + off);
 	}
@@ -436,25 +437,25 @@ heap_getsysattr(HeapTuple tup, int attnum, bool *isnull)
 			result = PointerGetDatum(&(tup->t_self));
 			break;
 		case ObjectIdAttributeNumber:
-			result = ObjectIdGetDatum(tup->t_data->t_oid);
+			result = ObjectIdGetDatum(HeapTupleGetOid(tup));
 			break;
 		case MinTransactionIdAttributeNumber:
-			result = TransactionIdGetDatum(tup->t_data->t_xmin);
+			result = TransactionIdGetDatum(HeapTupleHeaderGetXmin(tup->t_data));
 			break;
 		case MinCommandIdAttributeNumber:
-			result = CommandIdGetDatum(tup->t_data->t_cmin);
+			result = CommandIdGetDatum(HeapTupleHeaderGetCmin(tup->t_data));
 			break;
 		case MaxTransactionIdAttributeNumber:
-			result = TransactionIdGetDatum(tup->t_data->t_xmax);
+			result = TransactionIdGetDatum(HeapTupleHeaderGetXmax(tup->t_data));
 			break;
 		case MaxCommandIdAttributeNumber:
-			result = CommandIdGetDatum(tup->t_data->t_cmax);
+			result = CommandIdGetDatum(HeapTupleHeaderGetCmax(tup->t_data));
 			break;
 		case TableOidAttributeNumber:
 			result = ObjectIdGetDatum(tup->t_tableOid);
 			break;
 		default:
-			elog(ERROR, "heap_getsysattr: invalid attnum %d", attnum);
+			elog(ERROR, "invalid attnum: %d", attnum);
 			result = 0;			/* keep compiler quiet */
 			break;
 	}
@@ -511,41 +512,6 @@ heap_copytuple_with_tuple(HeapTuple src, HeapTuple dest)
 	memcpy((char *) dest->t_data, (char *) src->t_data, src->t_len);
 }
 
-#ifdef NOT_USED
-/* ----------------
- *		heap_deformtuple
- *
- *		the inverse of heap_formtuple (see below)
- * ----------------
- */
-void
-heap_deformtuple(HeapTuple tuple,
-				 TupleDesc tdesc,
-				 Datum *values,
-				 char *nulls)
-{
-	int			i;
-	int			natts;
-
-	Assert(HeapTupleIsValid(tuple));
-
-	natts = tuple->t_natts;
-	for (i = 0; i < natts; i++)
-	{
-		bool		isnull;
-
-		values[i] = heap_getattr(tuple,
-								 i + 1,
-								 tdesc,
-								 &isnull);
-		if (isnull)
-			nulls[i] = 'n';
-		else
-			nulls[i] = ' ';
-	}
-}
-#endif
-
 /* ----------------
  *		heap_formtuple
  *
@@ -571,18 +537,17 @@ heap_formtuple(TupleDesc tupleDescriptor,
 {
 	HeapTuple	tuple;			/* return tuple */
 	HeapTupleHeader td;			/* tuple data */
-	int			bitmaplen;
 	unsigned long len;
 	int			hoff;
 	bool		hasnull = false;
 	int			i;
 	int			numberOfAttributes = tupleDescriptor->natts;
 
-	if (numberOfAttributes > MaxHeapAttributeNumber)
-		elog(ERROR, "heap_formtuple: numberOfAttributes of %d > %d",
-			 numberOfAttributes, MaxHeapAttributeNumber);
-
-	len = offsetof(HeapTupleHeaderData, t_bits);
+	if (numberOfAttributes > MaxTupleAttributeNumber)
+		ereport(ERROR,
+				(errcode(ERRCODE_TOO_MANY_COLUMNS),
+				 errmsg("number of columns (%d) exceeds limit (%d)",
+						numberOfAttributes, MaxTupleAttributeNumber)));
 
 	for (i = 0; i < numberOfAttributes; i++)
 	{
@@ -593,13 +558,15 @@ heap_formtuple(TupleDesc tupleDescriptor,
 		}
 	}
 
-	if (hasnull)
-	{
-		bitmaplen = BITMAPLEN(numberOfAttributes);
-		len += bitmaplen;
-	}
+	len = offsetof(HeapTupleHeaderData, t_bits);
 
-	hoff = len = MAXALIGN(len); /* be conservative here */
+	if (hasnull)
+		len += BITMAPLEN(numberOfAttributes);
+
+	if (tupleDescriptor->tdhasoid)
+		len += sizeof(Oid);
+
+	hoff = len = MAXALIGN(len); /* align user data safely */
 
 	len += ComputeDataSize(tupleDescriptor, value, nulls);
 
@@ -615,15 +582,16 @@ heap_formtuple(TupleDesc tupleDescriptor,
 	td->t_natts = numberOfAttributes;
 	td->t_hoff = hoff;
 
-	DataFill((char *) td + td->t_hoff,
+	if (tupleDescriptor->tdhasoid)		/* else leave infomask = 0 */
+		td->t_infomask = HEAP_HASOID;
+
+	DataFill((char *) td + hoff,
 			 tupleDescriptor,
 			 value,
 			 nulls,
 			 &td->t_infomask,
 			 (hasnull ? td->t_bits : NULL));
 
-	td->t_infomask |= HEAP_XMAX_INVALID;
-
 	return tuple;
 }
 
@@ -647,7 +615,6 @@ heap_modifytuple(HeapTuple tuple,
 	char	   *nulls;
 	bool		isNull;
 	HeapTuple	newTuple;
-	uint8		infomask;
 
 	/*
 	 * sanity checks
@@ -664,14 +631,11 @@ heap_modifytuple(HeapTuple tuple,
 	 * allocate and fill *value and *nulls arrays from either the tuple or
 	 * the repl information, as appropriate.
 	 */
-	value = (Datum *) palloc(numberOfAttributes * sizeof *value);
-	nulls = (char *) palloc(numberOfAttributes * sizeof *nulls);
+	value = (Datum *) palloc(numberOfAttributes * sizeof(Datum));
+	nulls = (char *) palloc(numberOfAttributes * sizeof(char));
 
-	for (attoff = 0;
-		 attoff < numberOfAttributes;
-		 attoff += 1)
+	for (attoff = 0; attoff < numberOfAttributes; attoff++)
 	{
-
 		if (repl[attoff] == ' ')
 		{
 			value[attoff] = heap_getattr(tuple,
@@ -681,13 +645,13 @@ heap_modifytuple(HeapTuple tuple,
 			nulls[attoff] = (isNull) ? 'n' : ' ';
 
 		}
-		else if (repl[attoff] != 'r')
-			elog(ERROR, "heap_modifytuple: repl is \\%3d", repl[attoff]);
-		else
-		{						/* == 'r' */
+		else if (repl[attoff] == 'r')
+		{
 			value[attoff] = replValue[attoff];
 			nulls[attoff] = replNull[attoff];
 		}
+		else
+			elog(ERROR, "unrecognized replace flag: %d", (int) repl[attoff]);
 	}
 
 	/*
@@ -697,23 +661,109 @@ heap_modifytuple(HeapTuple tuple,
 							  value,
 							  nulls);
 
+	pfree(value);
+	pfree(nulls);
+
 	/*
-	 * copy the header except for t_len, t_natts, t_hoff, t_bits,
-	 * t_infomask
+	 * copy the identification info of the old tuple: t_ctid, t_self, and
+	 * OID (if any)
 	 */
-	infomask = newTuple->t_data->t_infomask;
-	memmove((char *) &newTuple->t_data->t_oid,	/* XXX */
-			(char *) &tuple->t_data->t_oid,
-			((char *) &tuple->t_data->t_hoff -
-			 (char *) &tuple->t_data->t_oid));	/* XXX */
-	newTuple->t_data->t_infomask = infomask;
-	newTuple->t_data->t_natts = numberOfAttributes;
+	newTuple->t_data->t_ctid = tuple->t_data->t_ctid;
 	newTuple->t_self = tuple->t_self;
 	newTuple->t_tableOid = tuple->t_tableOid;
+	if (relation->rd_rel->relhasoids)
+		HeapTupleSetOid(newTuple, HeapTupleGetOid(tuple));
 
 	return newTuple;
 }
 
+/* ----------------
+ *		heap_deformtuple
+ *
+ *		Given a tuple, extract data into values/nulls arrays; this is
+ *		the inverse of heap_formtuple.
+ *
+ *		Storage for the values/nulls arrays is provided by the caller;
+ *		it should be sized according to tupleDesc->natts not tuple->t_natts.
+ *
+ *		Note that for pass-by-reference datatypes, the pointer placed
+ *		in the Datum will point into the given tuple.
+ *
+ *		When all or most of a tuple's fields need to be extracted,
+ *		this routine will be significantly quicker than a loop around
+ *		heap_getattr; the loop will become O(N^2) as soon as any
+ *		noncacheable attribute offsets are involved.
+ * ----------------
+ */
+void
+heap_deformtuple(HeapTuple tuple,
+				 TupleDesc tupleDesc,
+				 Datum *values,
+				 char *nulls)
+{
+	HeapTupleHeader tup = tuple->t_data;
+	Form_pg_attribute *att = tupleDesc->attrs;
+	int			tdesc_natts = tupleDesc->natts;
+	int			natts;			/* number of atts to extract */
+	int			attnum;
+	char	   *tp;				/* ptr to tuple data */
+	long		off;			/* offset in tuple data */
+	bits8	   *bp = tup->t_bits;		/* ptr to null bitmask in tuple */
+	bool		slow = false;	/* can we use/set attcacheoff? */
+
+	natts = tup->t_natts;
+
+	/*
+	 * In inheritance situations, it is possible that the given tuple
+	 * actually has more fields than the caller is expecting.  Don't run
+	 * off the end of the caller's arrays.
+	 */
+	natts = Min(natts, tdesc_natts);
+
+	tp = (char *) tup + tup->t_hoff;
+
+	off = 0;
+
+	for (attnum = 0; attnum < natts; attnum++)
+	{
+		if (!HeapTupleNoNulls(tuple) && att_isnull(attnum, bp))
+		{
+			values[attnum] = (Datum) 0;
+			nulls[attnum] = 'n';
+			slow = true;		/* can't use attcacheoff anymore */
+			continue;
+		}
+
+		nulls[attnum] = ' ';
+
+		if (!slow && att[attnum]->attcacheoff >= 0)
+			off = att[attnum]->attcacheoff;
+		else
+		{
+			off = att_align(off, att[attnum]->attalign);
+
+			if (!slow)
+				att[attnum]->attcacheoff = off;
+		}
+
+		values[attnum] = fetchatt(att[attnum], tp + off);
+
+		off = att_addlength(off, att[attnum]->attlen, tp + off);
+
+		if (att[attnum]->attlen <= 0)
+			slow = true;		/* can't use attcacheoff anymore */
+	}
+
+	/*
+	 * If tuple doesn't have all the atts indicated by tupleDesc, read the
+	 * rest as null
+	 */
+	for (; attnum < tdesc_natts; attnum++)
+	{
+		values[attnum] = (Datum) 0;
+		nulls[attnum] = 'n';
+	}
+}
 
 /* ----------------
  *		heap_freetuple
@@ -736,12 +786,17 @@ heap_freetuple(HeapTuple htup)
  *
  * This routine forms a HeapTuple by copying the given structure (tuple
  * data) and adding a generic header.  Note that the tuple data is
- * presumed to contain no null fields.	It is typically only useful
- * for null-free system tables.
+ * presumed to contain no null fields and no varlena fields.
+ *
+ * This routine is really only useful for certain system tables that are
+ * known to be fixed-width and null-free.  It is used in some places for
+ * pg_class, but that is a gross hack (it only works because relacl can
+ * be omitted from the tuple entirely in those places).
  * ----------------
  */
 HeapTuple
 heap_addheader(int natts,		/* max domain index */
+			   bool withoid,	/* reserve space for oid */
 			   Size structlen,	/* its length */
 			   void *structure) /* pointer to the struct */
 {
@@ -753,7 +808,10 @@ heap_addheader(int natts,		/* max domain index */
 	AssertArg(natts > 0);
 
 	/* header needs no null bitmap */
-	hoff = MAXALIGN(offsetof(HeapTupleHeaderData, t_bits));
+	hoff = offsetof(HeapTupleHeaderData, t_bits);
+	if (withoid)
+		hoff += sizeof(Oid);
+	hoff = MAXALIGN(hoff);
 	len = hoff + structlen;
 
 	tuple = (HeapTuple) palloc(HEAPTUPLESIZE + len);
@@ -764,11 +822,13 @@ heap_addheader(int natts,		/* max domain index */
 	tuple->t_datamcxt = CurrentMemoryContext;
 	tuple->t_data = td = (HeapTupleHeader) ((char *) tuple + HEAPTUPLESIZE);
 
-	MemSet((char *) td, 0, len);
+	MemSet((char *) td, 0, hoff);
 
-	td->t_hoff = hoff;
 	td->t_natts = natts;
-	td->t_infomask = HEAP_XMAX_INVALID; /* XXX sufficient? */
+	td->t_hoff = hoff;
+
+	if (withoid)				/* else leave infomask = 0 */
+		td->t_infomask = HEAP_HASOID;
 
 	memcpy((char *) td + hoff, structure, structlen);
 
